@@ -58,6 +58,7 @@ class Enm:
 
         # define atoms
         native_pdb=cur_pdb.atom_slice(idxs)
+        
         coords=native_pdb.xyz[0]
         self.top=native_pdb.topology
         
@@ -72,6 +73,7 @@ class Enm:
         # find where distance is shorter than cutoff
         c_idx = (dmat<cutoff).nonzero()[0]
         m_idx = np.array(np.triu_indices(ll,1)).T[c_idx]
+
         # k = gamma/d^2
         k_elast=1./dmat[c_idx]**2
         
@@ -80,9 +82,9 @@ class Enm:
         if self.sparse:
             # construct matrix
             ele_up=np.zeros(len(c_idx)*9)
-            idx_up=np.zeros((2,len(c_idx)*9))
+            idx_up=np.zeros((2,len(c_idx)*9),dtype=np.int_)
             ele_diag=np.zeros(ll*9)
-            idx_diag=np.zeros((2,ll*9))
+            idx_diag=np.zeros((2,ll*9),dtype=np.int_)
             kkk=0
             for i in range(ll):
                 for mu in range(3):
@@ -103,9 +105,10 @@ class Enm:
                         # filling diagonal elements
                         ele_diag[9*i+3*mu+nu]+=temp
                         ele_diag[9*j+3*mu+nu]+=temp
-            idx_down=np.array((idx_up[1],idx_up[0]))
+            idx_down=np.array((idx_up[1],idx_up[0]),dtype=np.int_)
             idx_tot=np.concatenate([idx_up,idx_down,idx_diag],axis=1)
             ele_tot=np.concatenate([ele_up,ele_up,ele_diag])
+            
             mat=sp.csc_matrix((ele_tot,idx_tot))
             self.mat=mat # store the interaction matrix for future C2-C2 calculations
             # diagonalise
@@ -157,15 +160,23 @@ class Enm:
         """ Returns eigenvectors of the interaction matrix of the ENM"""
         return self.e_vec
 
-    def get_MSF(self):
+    def get_MSF(self,pdb_file=None):
         """ Return mean squared fluctuations of each bead"""
         e_vec=self.get_evec()
         e_val=self.get_eval()
         cacca=np.sum(e_vec.reshape(self.n_beads,3,e_vec.shape[1])**2,axis=1)
         msf=np.sum(cacca[:,self.n_null_modes:]*(1/e_val[self.n_null_modes:]),axis=1)
+
+        if(pdb_file!=None):
+            pdb=md.Trajectory(self.coords,self.top)
+            pdb.save_pdb(pdb_file,bfactors=msf)
         return msf
     
-    
+    def get_beads(self):
+        #rr = ["%s_%s_%d_%d" % (at.name,at.residue.resn,at.residue,at.chainid) for at in self.top.atoms]
+        rr = ["%s_%s_%d_%s" % (at.name,at.residue.name,at.residue.resSeq,at.residue.segment_id) for at in self.top.atoms]
+        return rr
+
     
     def c2_fluctuations(self):
         """ Computes the C2-C2 fluctuations of an RNA ENM.
