@@ -18,6 +18,7 @@ from builtins import range
 import sys
 import mdtraj as md
 from scipy.spatial import distance
+
 import itertools
 import numpy as np
 import os
@@ -25,7 +26,7 @@ from . import definitions
 from . import nucleic
 from . import calc_mats as ff
     
-def ermsd(reference,target,cutoff=2.4,topology=None):
+def ermsd(reference,target,cutoff=2.4,topology=None,residues_ref=None,residues_target=None):
     
     """
     Calculate ermsd between reference and target structures  
@@ -59,9 +60,9 @@ def ermsd(reference,target,cutoff=2.4,topology=None):
     warn += "# Loaded target %s \n" % target
     sys.stderr.write(warn)
 
-    return ermsd_traj(ref,traj,cutoff=cutoff)
+    return ermsd_traj(ref,traj,cutoff=cutoff,residues_ref=residues_ref,residues_target=residues_target)
 
-def ermsd_traj(reference,traj,cutoff=2.4):
+def ermsd_traj(reference,traj,cutoff=2.4,residues_ref=None,residues_target=None):
     
     top_traj = traj.topology
     # initialize nucleic class
@@ -71,16 +72,25 @@ def ermsd_traj(reference,traj,cutoff=2.4):
     # initialize nucleic class
     nn_ref = nucleic.Nucleic(top_ref)
 
-    assert(len(nn_traj.ok_residues)==len(nn_ref.ok_residues))
     
-    coords_ref = reference.xyz[0,nn_ref.indeces_lcs]
+    ref_indeces = nn_ref.indeces_lcs
+    if(residues_ref!=None): ref_indeces = ref_indeces[:,residues_ref]
+        
+    coords_ref = reference.xyz[0,ref_indeces]
     ref_mat = ff.calc_gmat(coords_ref,cutoff).reshape(-1)
+
     #rna_seq = ["%s_%s_%s" % (res.name,res.resSeq,res.chain.index) for res in nn.ok_residues]
     gmats = []
+    traj_indeces = nn_traj.indeces_lcs
+    if(residues_target!=None): traj_indeces = traj_indeces[:,residues_target]
+    #print(residues_target)
+    assert ref_indeces.shape[1]==traj_indeces.shape[1], "reference %s, target %s" % (ref_indeces.shape,traj_indeces.shape)
+    
     for i in range(traj.n_frames):
-        coords_lcs = traj.xyz[i,nn_traj.indeces_lcs]
+        coords_lcs = traj.xyz[i,traj_indeces]
         gmats.append(ff.calc_gmat(coords_lcs,cutoff).reshape(-1))
-    dd = distance.cdist([ref_mat],gmats)/np.sqrt(len(nn_traj.ok_residues))
+    dd = distance.cdist([ref_mat],gmats)/np.sqrt(traj_indeces.shape[1])
+
     return dd[0]
 
 ############## ERMSD ###############
@@ -176,7 +186,7 @@ def dump_gvec_traj(traj,cutoff=2.4):
 #################################################
 
 
-def rmsd(reference,target,topology=None,out=None):
+def rmsd(reference,target,topology=None,out=None,heavy_atom=False):
     
     """
     Calculate rmsd after optimal alignment between reference and target structures. Superposition and RMSD calculations are performed using all heavy atoms. 
@@ -192,6 +202,9 @@ def rmsd(reference,target,topology=None,out=None):
          Topology filename. Must be specified if target is a trajectory.
     out :  string, optional
          If a string is specified, superimposed PDB structures are written to disk with the specified prefix.
+    heavy_atom :   bool, optional
+         If True, all heavy atoms are used for superposition. Default is False
+
     Returns
     -------
     array :
@@ -208,10 +221,10 @@ def rmsd(reference,target,topology=None,out=None):
         traj = md.load(target,top=topology)
     warn += "# Loaded target %s \n" % target
 
-    return rmsd_traj(ref,traj,out=out)
+    return rmsd_traj(ref,traj,out=out,heavy_atom=heavy_atom)
 
 
-def rmsd_traj(reference,traj,out=None):
+def rmsd_traj(reference,traj,out=None,heavy_atom=False):
     
     top_traj = traj.topology
     # initialize nucleic class
@@ -220,8 +233,13 @@ def rmsd_traj(reference,traj,out=None):
     top_ref = reference.topology
     # initialize nucleic class
     nn_ref = nucleic.Nucleic(top_ref)
-    assert(len(nn_traj.ok_residues)==len(nn_ref.ok_residues))
 
+    assert(len(nn_traj.ok_residues)==len(nn_ref.ok_residues))    
+    # check that sequence is identical when using heavy atoms
+    if(nn_traj.rna_seq_id!=nn_ref.rna_seq_id and heavy_atom==True):
+        sys.stderr.write("# Sequences are not identical, cannot superimpose using all heavy atoms. ")
+        sys.exit(1)
+        
     # loop over residues and find common heavy atoms
     idx_ref = []
     idx_target = []
@@ -234,16 +252,13 @@ def rmsd_traj(reference,traj,out=None):
         resname1 = nn_ref.rna_seq_id[ii]
         resname2 = nn_traj.rna_seq_id[ii]
         
-        # if the nucleotide is the same, use all atoms
-        if(res1.name == res2.name):
+        # if heavy_atom is true, use all atoms
+        if(heavy_atom):
             
             name2 = [at.name for at in res2.atoms if  at.name in definitions.nt_atoms[resname2]]
-            #name1 = [at.name for at in res1.atoms if  at.name in definitions.nb_atoms[res1.name]]
-            #print name2,name1
             for at in res1.atoms:
                 if at.name in definitions.nt_atoms[resname1]:
                     if(at.name in name2):
-                        #print at.name
                         idx_ref.append(at.index)
                         idx_target.append(((res2.atom(at.name)).index))
         # else, use bb only
@@ -254,7 +269,6 @@ def rmsd_traj(reference,traj,out=None):
                     if(at.name in name2):
                         idx_ref.append(at.index)
                         idx_target.append(((res2.atom(at.name)).index))
-            
     print("# found ",len(idx_ref), "atoms in common")
     
     if(len(idx_ref)<3):
@@ -263,8 +277,8 @@ def rmsd_traj(reference,traj,out=None):
         sys.exit(1)
         
     traj.superpose(reference,atom_indices=idx_target, ref_atom_indices=idx_ref)
-    if(out!=None):
-        traj.save(out)
+    if(out!=None): traj.save(out)
+    
     rmsd = np.sqrt(3*np.mean((traj.xyz[:, idx_target, :] - reference.xyz[0,idx_ref, :])**2, axis=(1,2)))
     return rmsd
 
@@ -406,7 +420,7 @@ def sugar_angles_traj(traj,residues=None,angles=None):
 
 #############################################################
 
-def pucker_angles(filename,topology=None,residues=None):
+def pucker_angles(filename,topology=None,residues=None,altona=False):
     
     """
     Calculate sugar pucker pseudorotation  torsion angles: phase and amplitude
@@ -435,9 +449,12 @@ def pucker_angles(filename,topology=None,residues=None):
         traj = md.load(filename,top=topology)
     warn = "# Loading %s \n" % filename
     sys.stderr.write(warn)
-    return pucker_angles_traj(traj,residues=residues)
-
-def pucker_angles_traj(traj,residues=None):
+    if(altona):
+        return pucker_altona_traj(traj,residues=residues)
+    else:
+        return pucker_rao_traj(traj,residues=residues)
+        
+def pucker_altona_traj(traj,residues=None):
 
     torsions,rr = sugar_angles_traj(traj,residues=residues)
     x1 = torsions[:,:,4] +  torsions[:,:,1] -  torsions[:,:,3] -   torsions[:,:,0]
@@ -448,6 +465,22 @@ def pucker_angles_traj(traj,residues=None):
     angles = np.dstack((phase,tm))
     return angles, rr
 
+def pucker_rao_traj(traj,residues=None):
+
+    torsions,rr = sugar_angles_traj(traj,residues=residues)
+
+    period = 4.*np.pi/5
+    A = (2./5.)*(torsions[:,:,0]+np.cos(period)*torsions[:,:,1] + np.cos(2.*period)*torsions[:,:,2] + \
+                np.cos(3.*period)*torsions[:,:,3] + np.cos(4.*period)*torsions[:,:,4])
+    B = -(2./5.)*(np.sin(period)*torsions[:,:,1] + np.sin(2.*period)*torsions[:,:,2] + \
+                np.sin(3.*period)*torsions[:,:,3] + np.sin(4.*period)*torsions[:,:,4])
+    
+    phase = np.arctan2(B,A) - 0.5*period
+    phase[np.where(phase<0.0)] += 2.0*np.pi
+    tm = np.sqrt(A*A + B*B)
+    angles = np.dstack((phase,tm))
+    return angles, rr
+
 ################################################################
 
 ################################################################
@@ -511,7 +544,7 @@ def jcouplings_traj(traj,residues=None,couplings=None,raw=False):
                 idx_angles1.append(i)
             else:
                 msg = "# Fatal error. requested coupling \"%s\" not available.\n" % couplings[i]
-                msg += "# Choose from: %s \n" % couplings_idx
+                msg += "# Choose from: %s \n" % definitions.couplings_idx
                 sys.stderr.write(msg)
                 sys.exit(1)
    
@@ -656,14 +689,19 @@ def ss_motif_traj(ref,traj,threshold=0.8,cutoff=2.4,sequence=None,bulges=0,out=N
                 # align whatever is in common in the backbone
                 idx_target = []
                 idx_ref = []
-                for res1,res2 in zip(nn_ref.ok_residues,traj_slice.topology.residues):
+                #for res1,res2 in zip(nn_ref.ok_residues,traj_slice.topology.residues):
+                for ii in range(len(nn_ref.ok_residues)):
+                    res1 = nn_ref.ok_residues[ii]
+                    res2 = list(traj_slice.topology.residues)[ii]
+                    
                     name2 = [at.name for at in res2.atoms if  at.name in definitions.bb_atoms]
                     for at in res1.atoms:
                         if at.name in definitions.bb_atoms:
                             if(at.name in name2):
                                 idx_ref.append(at.index)
                                 idx_target.append(((res2.atom(at.name)).index))
-                                #idx_target.append(res2[name2.index(at.name)].index)
+                                #print(at,res2.atom(at.name))
+
                 traj_slice.superpose(ref,atom_indices=idx_target, ref_atom_indices=idx_ref)
                 traj_slice.save(pdb_out)
 
@@ -718,7 +756,7 @@ def ds_motif(query,target,l1,l2,threshold=0.9,cutoff=2.4,topology=None,sequence=
     warn += "# Loaded target %s \n" % target
     sys.stderr.write(warn)
 
-    return ds_motif_traj(ref,traj,l1,l2,threshold=threshold,sequence=sequence,bulges=bulges,out=out)
+    return ds_motif_traj(ref,traj,l1,l2,cutoff=cutoff,threshold=threshold,sequence=sequence,bulges=bulges,out=out)
 
 def ds_motif_traj(ref,traj,l1,l2,threshold=0.9,cutoff=2.4,sequence=None,bulges=0,out=None):
     
@@ -812,11 +850,11 @@ def ds_motif_traj(ref,traj,l1,l2,threshold=0.9,cutoff=2.4,sequence=None,bulges=0
             resname_idxs = [nn_traj.rna_seq[l] for l  in idxs_combo[k]]
             
             results.append([i,dd_combo[0,k]/np.sqrt(l1 + l2),resname_idxs])
-
             #print results[-1]
             # Write aligned PDB 
             if(out != None):
-                pdb_out = "%s_%05d_%s_%d.pdb" % (out,count,resname_idxs[k][0],i)
+
+                pdb_out = "%s_%05d_%s_%d.pdb" % (out,count,resname_idxs[0],i)
                 # slice trajectory
                 tmp_atoms = []
                 tmp_res =[]
@@ -828,7 +866,10 @@ def ds_motif_traj(ref,traj,l1,l2,threshold=0.9,cutoff=2.4,sequence=None,bulges=0
                 # align whatever is in common in the backbone
                 idx_target = []
                 idx_ref = []
-                for res1,res2 in zip(nn_ref.ok_residues,traj_slice.topology.residues):
+                for ii in range(len(nn_ref.ok_residues)):
+                    res1 = nn_ref.ok_residues[ii]
+                    res2 = list(traj_slice.topology.residues)[ii]
+                    
                     name2 = [at.name for at in res2.atoms if  at.name in definitions.bb_atoms]
                     for at in res1.atoms:
                         if at.name in definitions.bb_atoms:
@@ -847,8 +888,8 @@ def ds_motif_traj(ref,traj,l1,l2,threshold=0.9,cutoff=2.4,sequence=None,bulges=0
 
 
 
-def annotate(filename,topology=None):
-    
+def annotate(filename,topology=None, stacking_rho_cutoff=2.5, stacking_angle_cutoff=40, pairing_angle_cutoff=60):
+
     """
     Find base-pair and base-stacking 
 
@@ -858,6 +899,12 @@ def annotate(filename,topology=None):
          Filename of structure, any format accepted by MDtraj can be used.
     topology : string, optional
          Topology filename. Must be specified if target is a trajectory.
+    stacking_rho_cutoff : float
+         Cutoff for base stacking. Rho distance. Default is 2.5 (unit in angstroms).
+    stacking_angle_cutoff : float
+         Cutoff for base stacking. Angle between the vectors normal to the planes of the two bases. Default is 40 (unit in degrees).
+    stacking_angle_cutoff : float
+         Cutoff for base pairing. Angle between the vectors normal to the planes of the two bases. Default is 60 (unit in degrees).
     Returns
     -------
     stackings : list
@@ -876,11 +923,11 @@ def annotate(filename,topology=None):
     warn = "# Loading %s \n" % filename
     sys.stderr.write(warn)
     
-    return annotate_traj(traj)
+    return annotate_traj(traj, stacking_rho_cutoff, stacking_angle_cutoff, pairing_angle_cutoff)
 
 
-def annotate_traj(traj):
-
+def annotate_traj(traj, stacking_rho_cutoff=2.5, stacking_angle_cutoff=40, pairing_angle_cutoff=60):
+ 
     # this is the binning for annotation
     bins = [0,1.84,3.84,2.*np.pi]
     bins_label = ["W","H","S"]
@@ -917,16 +964,19 @@ def annotate_traj(traj):
         # find stacked bases 
 
         # z_ij  AND z_ji > 2 AA
-        stackz_12 = np.where(z_12>0.04)
-        stackz_21 = np.where(z_21>0.04)
+        stackz_12 = np.where(z_12>0.04)  # nm^2
+        stackz_21 = np.where(z_21>0.04)  # nm^2
         
         # rho_ij OR rho_ji < 2.5 AA
-        rhoz_12 =  np.where(rho_12<0.0625)
-        rhoz_21 =  np.where(rho_21<0.0625)
+        #rhoz_12 =  np.where(rho_12<0.0625)
+        #rhoz_21 =  np.where(rho_21<0.0625)
+        rhoz_12 =  np.where(rho_12<stacking_rho_cutoff * stacking_rho_cutoff * 0.1 * 0.1)  # nm^2
+        rhoz_21 =  np.where(rho_21<stacking_rho_cutoff * stacking_rho_cutoff * 0.1 * 0.1)  # nm^2
         union_rho = np.union1d(rhoz_12[0],rhoz_21[0])
 
         # angle between normal planes < 40 deg
-        anglez = np.where(np.abs(angles) >0.766)
+        #anglez = np.where(np.abs(angles) >0.766)
+        anglez = np.where(np.abs(angles) > np.cos(np.radians(stacking_angle_cutoff)))   # cosine of angle beween the normal vectors constructed on base i and base j
 
         # intersect all criteria
         inter_zeta = np.intersect1d(stackz_12,stackz_21)
@@ -975,7 +1025,8 @@ def annotate_traj(traj):
         for j in range(len(paired_pairs)):
 
             # if angle is larger than 60 deg, skip
-            if(np.abs(angles[paired[j]]) < 0.5 ): continue
+            #if(np.abs(angles[paired[j]]) < 0.5 ): continue
+            if(np.abs(angles[paired[j]]) < np.cos(np.radians(pairing_angle_cutoff))): continue
 
             index1 = paired_pairs[j][0]
             index2 = paired_pairs[j][1]
@@ -1082,8 +1133,14 @@ def dot_bracket(pairings,sequence):
                 idx2 = pp[0][e][1]
 
                 # check that bracket is not there already. This might happen in simulations
-                if(string[idx1]!="." or string[idx2]!="." ):
-                    warn = "# Frame %d, residue %s has a double WC base-pairs. " % (k,sequence[idx1],sequence[idx2])
+                if(string[idx1]!="."):
+                    warn = "# Frame %d, residue %s has a double WC base-pairs. " % (k,sequence[idx1])
+                    warn += " Dot-bracket annotation set to XXX \n"
+                    sys.stderr.write(warn)
+                    dot_bracket.append("XXX")                            
+                    continue
+                if(string[idx2]!="." ):
+                    warn = "# Frame %d, residue %s has a double WC base-pairs. " % (k,sequence[idx2])
                     warn += " Dot-bracket annotation set to XXX \n"
                     sys.stderr.write(warn)
                     dot_bracket.append("XXX")                            
@@ -1184,5 +1241,385 @@ def snippet(pdb,sequence,outdir=None):
          
             ii += 1
         idx = cur_pdb.read()
-      
+
+
+############# SEC_STRUCTURE ########################################
+
+def parse_dotbr(dotbra):
+    res_bp = []
+    basepairs = []
+    for k, i in enumerate(dotbra):
+        if i in [")", "]", "}"]:
+            if i == ")":
+                search = "("
+            elif i == "]":
+                search = "["
+            else:
+                search = "{"
+
+            res_bp.append(k)
+            for k2 in range(k-1)[::-1]:
+                j = dotbra[k2]
+                if j == search and k2 not in res_bp:
+                    basepairs.append([k2, k])
+                    res_bp.append(k2)
+                    break
+    return basepairs            
     
+def parse_dotbracket(threshold, file, weights):
+    import barnaba.sec_str_constants as secon
+    import re 
+    print("Parsing file ", file)    
+    with open(file) as f:
+        f_dotbracket = f.readlines()
+        f.close()
+    pdbs = []
+    ann_lists = []
+    ann_list = {}
+    res_bp = []
+    n_frames = 0
+    n_pdbs = 0
+    traj = False
+    list_base_pairs = []
+    sequence = []
+    mult_chain = False
+    for line in f_dotbracket:
+        if line.startswith("#") and line.split()[1] == "sequence":
+            sequence = np.array([r.split("_") for r in line.split()[2].split("-")])
+            if len(np.unique(sequence[:,2])) > 1:
+                print("Found multiple chains, drawing first chain")
+          #      sequence = sequence[np.where(sequence[:,2]==np.unique(sequence[:,2])[0])]
+          #      print(sequence)
+          #      mult_chain = True
+        if not line.startswith("#"):
+            l = line.split()
+            if "pdb" in l[0].lower():
+                dotbr = l[1]
+                if len(dotbr) != len(sequence):
+                    print(mult_chain)
+                    if not mult_chain:
+                        sys.exit("Dot-bracket length does not match sequence length")
+                    else:
+                        dotbr = dotbr[:len(sequence)]
+                base_pairs = parse_dotbr(dotbr)
+                for bp in base_pairs:
+                    ann_list[bp[0], bp[1], "WCc"] = 1.
+                list_base_pairs = base_pairs    
+                ann_lists.append(dict(ann_list))
+                ann_list = {}
+            else:
+                try: int(l[0])
+                except ValueError:
+                    continue
+                else:
+                    traj = True
+                    dotbr = l[1]
+                    if len(dotbr) != len(sequence):
+                        if not mult_chain:
+                            sys.exit("Dot-bracket length does not match sequence length")    
+                        else:
+                            dotbr = dotbr[:len(sequence)]
+                    n_frames += 1
+                    base_pairs = parse_dotbr(dotbr)
+                    for bp in base_pairs:
+                        if bp not in list_base_pairs:
+                            list_base_pairs.append(bp)
+                        try:
+                            ann_list[bp[0], bp[1], "WCc"]
+                        except:
+                            if len(weights) == 0:
+                                ann_list[bp[0], bp[1], "WCc"] = 1.
+                            else:
+                                ann_list[bp[0], bp[1], "WCc"] = weights[n_frames-1]
+                        else: 
+                            if len(weights) == 0:
+                                ann_list[bp[0], bp[1], "WCc"] += 1.
+                            else:    
+                                ann_list[bp[0], bp[1], "WCc"] += weights[n_frames-1]
+    if len(list_base_pairs) == 0:
+        sys.exit("No basepairs found.")
+    if traj:
+        for ann, value in ann_list.items():
+            if len(weights) == 0:
+                ann_list[ann] /= n_frames
+            else:
+                ann_list[ann] /= sum(weights)
+    chains = [0]
+    return sequence, ann_list, list_base_pairs, n_frames    
+
+def parse_annotations(threshold, file, weights):
+    import barnaba.sec_str_constants as secon
+    import re 
+    print("Parsing file ", file)    
+    with open(file) as f:
+        annotation = f.readlines()
+        f.close()
+    
+    re_pdb = re.compile("^\s*#\s+PDB\s+.*$")
+    re_frame = re.compile("^\s*#\s+Frame.*$")
+    re_ann = re.compile("^\s*([AGCU])_([0-9]+)_([0-9]+)\s+([AGCU])_([0-9]+)_([0-9]+)\s+.*$")
+    chains = []
+    old_C = "X"
+    ann_lists = []
+    ann_list = {}
+    n_frames = 0
+    for line in annotation:
+        if line.startswith("#") and line.split()[1] == "sequence":
+            sequence = np.array([r.split("_") for r in line.split()[2].split("-")])
+        pdb = re_pdb.match(line)
+        if pdb:
+            n_frames += 1
+            continue
+        frame = re_frame.match(line)
+        if frame:
+            n_frames += 1
+            continue
+        ann = re_ann.match(line)
+        if ann:
+            cols = line.split()
+            i_N = cols[0].split("_")[0]
+            ri = int(cols[0].split("_")[1])
+            i_C = cols[0].split("_")[2]
+            j_N = cols[1].split("_")[0]
+            rj = int(cols[1].split("_")[1])
+            j_C = cols[1].split("_")[2]
+            ann_ij = cols[2]
+            if ann_ij == "XXX":
+                continue
+            i = np.where(np.array(sequence)[:,1].astype(int) == ri)[0][0]
+            j = np.where(np.array(sequence)[:,1].astype(int) == rj)[0][0]
+            if old_C != "X" and (i_C != j_C or i_C != old_C):
+                ann_lists.append(ann_list)
+                ann_list = {}
+            if i_C not in chains:
+                chains.append(i_C)
+            old_C = i_C
+            try:
+                ann_list[i, j, ann_ij]
+            except:
+                if len(weights) == 0:
+                    ann_list[i, j, ann_ij] = 1.
+                else:    
+                    ann_list[i, j, ann_ij] = weights[n_frames-1]
+            else: 
+                if len(weights) == 0:
+                    ann_list[i, j, ann_ij] += 1.
+                else:    
+                    ann_list[i, j, ann_ij] += weights[n_frames-1]
+        
+    pairs = []
+    for ann, value in ann_list.items():
+        if len(weights) == 0:
+            ann_list[ann] /= n_frames
+        else:
+            ann_list[ann] /= sum(weights)
+        if ann_list[ann] >= threshold and [ann[0], ann[1]] not in pairs:
+            pairs.append([ann[0], ann[1]])
+    return sequence, ann_list, pairs, n_frames
+    
+def stems(param_wc, param_bp, param_stack):
+    
+    #pairs_wc =  np.unique(np.array(param_wc)[:,1:3], axis=0)
+    #pairs_all = np.unique(np.array(param_bp + param_stack + param_wc)[:,1:3], axis=0)
+    pairs_wc = definitions.unique_rows(np.array(param_wc)[:,1:3])
+    pairs_all = definitions.unique_rows(np.array(param_bp + param_stack + param_wc)[:,1:3])
+    sums_all = np.unique( np.sum(pairs_all, axis=1))
+    ds_all = []
+   
+    for p in pairs_all:
+        added = False
+        for di, ds in enumerate(ds_all):
+            if abs(ds[-1,0]-p[0])==1. and abs(ds[-1,1]-p[1])==1.:
+                if len(ds) > 1:
+                    if ds[-1,0]-p[0] == ds[-2,0]-ds[-1,0] and ds[-1,1]-p[1] == ds[-2,1]-ds[-1,1]:
+                        added = True
+                        ds_all[di] = np.append(ds, [p], axis=0)
+                        break
+                else:
+                    ds_all[di] = np.append(ds, [p], axis=0)
+                    added = True
+                    break
+        if not added:
+            ds_all.append(np.array([p]))
+
+    wc_per_ds = [[p for p in s if tuple(p) in set(tuple(x) for x in pairs_wc)] for s in ds_all]
+    # sort double strands first by number of wc pairs, then number of all pairs
+    sortkeys = np.lexsort((np.array([len(st) for st in ds_all]), np.array([len(wc) for wc in wc_per_ds])))
+    ds_all = np.array(ds_all)[sortkeys][::-1]
+    stems = []
+    diagonal_pairs = []
+    lonely_pairs = np.array([])
+    for k, d in enumerate(ds_all):
+        limits = np.array([[d[:,0].min(), d[:,1].max()], [d[:,0].max(), d[:,1].min()]])
+        add = True
+        for d2 in ds_all[:k]:
+            l2 = np.array([[d2[:,0].min(), d2[:,1].max()], [d2[:,0].max(), d2[:,1].min()]])
+            if ((l2[0][0] <= limits[0][0] <= l2[1][0] and l2[1][1] <= limits[0][1] <= l2[0][1]) or
+                    (l2[0][0] <= limits[1][0] <= l2[1][0] and l2[1][1] <= limits[1][1] <= l2[0][1])):
+                add = False
+                diagonal_pairs += list(d)        
+        if add:
+            if len(d) > 1:
+                stems.append(d)
+            else:
+                if len(lonely_pairs) == 0:
+                    lonely_pairs = np.array(d)
+                else:    
+                    lonely_pairs = np.append(lonely_pairs, np.array(d), axis=0)
+    return stems, diagonal_pairs, lonely_pairs
+
+def parameters(pairs, ann_list, n, threshold):
+    import barnaba.sec_str_constants as secon
+
+    # parameters:
+    # potential type 0: harmonic potential
+    # potential type 1: semiharmonic potential to reject bases
+    # potential type 2: angular potential to keep ds regions in order
+    # potential type 3: semiharmonic potential to reject terminal bases in x direction
+  
+
+    param_seq = np.empty((0, 5))
+    for i in range(n-1):
+        param_seq = np.append(param_seq, [[0, i, i+1, secon.k_seq, secon.d_seq]], axis=0)
+    param_bp = []
+    param_wc = []
+    param_stack = []
+    for ann, value in ann_list.items():
+        i = ann[0]
+        j = ann[1]
+        ann_ij = ann[2]
+        if value < threshold or abs(i-j) < 2:
+            continue
+        if ann_ij in secon.list_bp_ct:
+            if ann_ij in secon.list_wc_pairs: 
+                param_wc.append([0, i, j, secon.k_wc*value, secon.d_short])
+            else:
+                param_bp.append([0, i, j, secon.k_bp*value, secon.d_short])
+        if ann_ij in secon.list_stackings:
+            if abs(i-j) == 2:    
+                param_stack.append([0, i, j, secon.k_seq*value, secon.d_stack])
+            else:
+                param_stack.append([0, i, j, secon.k_stack*value, secon.d_short])
+    pairs_stems, diagonal_pairs, lonely_pairs = stems(param_wc, param_bp, param_stack)
+    param_wc_ds = []
+ #   print("pairs_stems", pairs_stems)
+ #   print("diagonal_pairs", diagonal_pairs)
+ #   print("lonely_pairs", lonely_pairs)
+    # Longest stem vertical
+    param_stem = np.empty((0, 4))
+    if len(pairs_stems) > 0:
+        diff = np.absolute(pairs_stems[0][:,1]-pairs_stems[0][:,0])
+        keys = np.argsort(diff)
+        for k in keys:
+            pair = pairs_stems[0][k]
+            if abs(pair[0]-pair[1]) > 2:
+                param_stem = np.append(param_stem, [[4, int(pair[0:2].min()), int(pair[0:2].max()), 0]], axis=0)
+    else:
+        if len(lonely_pairs) > 0:
+            for pair in lonely_pairs:
+                if sum(pair) == sum(lonely_pairs[np.argsort(lonely_pairs[:,0])[0]]):
+                    param_stem = np.append(param_stem, [[4, int(pair[0:2].min()), int(pair[0:2].max()), 0]], axis=0)
+
+    for pair in diagonal_pairs:
+        for pi in param_wc + param_bp + param_stack:
+            if tuple(pi[1:3]) in [tuple(pair), tuple(pair[::-1])]:
+                pi[3] = 0
+                pi[4] = secon.d_long
+
+    param_ds = []
+    param_ang = np.empty((0, 7))
+    if len(lonely_pairs) > 0:
+        all_stems = pairs_stems
+        for p in lonely_pairs:
+            all_stems.append(np.array([p]))
+    else:
+        all_stems = pairs_stems
+    for k, stem in enumerate(all_stems):
+        stem_limits = np.array([stem[:,0].min(), stem[:,1].max(), stem[:,0].max(), stem[:,1].min()]).astype(int)
+        same_stem = stem.copy()
+        for k2, s in enumerate(all_stems):
+            if k!=k2:
+                if sum(s[0]) == sum(stem[0]):
+                    same_stem = np.append(same_stem, s, axis=0)
+                elif abs(sum(s[0]) - sum(stem[0])) == 1:    
+                    s_limits = np.array([s[:,0].min(), s[:,1].max(), s[:,0].max(), s[:,1].min()]).astype(int)
+                     
+                    if len(np.unique(np.append(stem_limits, s_limits))) == len(np.unique(s_limits)) + len(np.unique(stem_limits)):
+                        if (abs(stem_limits[0]-s_limits[2]) == 1 or abs(stem_limits[1]-s_limits[3]) == 1 or
+                            abs(s_limits[0]-stem_limits[2]) == 1 or abs(s_limits[1]-stem_limits[3]) == 1):
+                            same_stem = np.append(same_stem, s, axis=0)
+        same_stem = same_stem[np.argsort(same_stem[:,0] ),:]                    
+        limits = np.array([same_stem[:,0].min(), same_stem[:,1].max(), same_stem[:,0].max(), same_stem[:,1].min()]).astype(int)
+        n_stem = len(same_stem)
+        param_ds = []
+        t_stems = [tuple(s) for s in stem]
+
+        for t_pair in t_stems:
+            for pi in param_bp + param_stack + param_wc:
+                if tuple(pi[1:3]) == t_pair or tuple(pi[1:3][::-1]) == t_pair:
+                    pi[3] *= 1.2**(n_stem-1)
+            p1 = t_pair[0]       
+            p2 = t_pair[1]
+            key, dim  = np.where(same_stem==t_pair)
+            key = key[0]
+            if abs(p1-p2) > 2:
+                if key < len(same_stem)-1:
+                    pul = same_stem[key+1][0]
+                    pur = same_stem[key+1][1]
+                    factor = 0.8**(abs(p1-pul)+abs(p2-pur)-2)
+                    param_ang = np.append(param_ang, [[2, p1, p2, pul, pur, factor*n_stem * secon.k_ang, secon.angle]], axis=0)
+        for pi in param_wc:
+            if pi[1:3] in stem or pi[1:3][::-1] in stem:
+                param_ds.append(pi)
+        param_wc_ds.append(param_ds)
+
+    sorted_params = np.empty((0,5))
+    n_excl = 0
+    for i in range(n):
+        i_list = []
+        for p in param_wc+param_bp+param_stack:
+            if i == max(p[1:3]) and p[3] == 0:
+                n_excl += 1
+                continue
+    #        elif i == max(p[1:3]) and abs(p[1]-p[2])<2:
+    #            n_excl += 1
+            elif i == max(p[1:3]):
+                i_list.append(p)
+        if len(i_list) == 0:
+            continue
+        diff = np.array([abs(p[1]-p[2]) for p in i_list])
+        keys = np.argsort(diff)
+        for k in keys:
+            sorted_params = np.append(sorted_params, [i_list[k]], axis=0)
+    assert (len(sorted_params)+n_excl == len(param_wc+param_bp+param_stack))        
+
+    param_angle_180 = np.empty((0, 6))
+    param_bulge = np.empty((0, 6))
+    param_bulge_rep = np.empty((0, 6))
+    paired = True
+    pair_now = []
+    for i in range(1, n-1):
+        param_angle_180 = np.append(param_angle_180, [[8, i-1, i, i+1, secon.k_angle_straight, 0]], axis=0)        
+    
+    param_rep = np.empty((0, 5))
+    param_rep_lr = np.empty((0, 5))
+    i1, i2 = np.triu_indices(n, k=1)
+    n_ij = i1.shape[0]
+    _type = np.full((n_ij), 1)
+    _k = np.full((n_ij), secon.k_rep2)
+    _d = np.full((n_ij), secon.d_rep2)
+    param_rep = np.append(param_rep, np.column_stack((_type, i1, i2, _k, _d)), axis=0)
+    _type = np.full((n_ij), 7)
+    _k = np.full((n_ij), secon.k_rep_lr)
+    _d = np.full((n_ij), 0)
+    param_rep_lr = np.append(param_rep_lr, np.column_stack((_type, i1, i2, _k, _d)), axis=0)
+
+    return param_seq, param_rep, param_rep_lr, sorted_params, param_stem, param_ang, param_bulge, param_bulge_rep, param_angle_180   
+
+def get_par(sorted_params, pos):
+    i1 = sorted_params[:,1].astype(int)
+    i2 = sorted_params[:,2].astype(int)
+    d = np.linalg.norm(pos[i1]-pos[i2], axis=1)
+
+    ii = np.argmin(d)
+    return sorted_params[ii], ii
